"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the execution time logic.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from collections import OrderedDict
from datetime import datetime

from colorama import Fore, Style

import PyFunceble


class ExecutionTime:  # pylint: disable=too-few-public-methods
    """
    Sets and returns the execution time of the program.

    :param str action:
        The action related the execution time.
        Can be `start` or `stop`.

    :param bool last:
        Tell the subsystem if we are at the very end of the file testing.
    """

    authorized = False

    def __init__(self, action="start", last=False):
        self.authorized = self.authorization()

        # We parse the action to the class scope.
        self.action = action

        if self.action == "start":
            # The action is equal to `start`.

            # We set the starting time.
            self.set_starting_time()
        elif self.action == "stop":
            # * The action is not equal to `start`.
            # and
            # * The action is equal to `stop`

            # We set the ending time.
            self.set_stoping_time()

        if self.authorized and self.action == "stop":
            # * We are authorized to operate.
            # and
            # * The given action is stop.

            formatted_time = self.format_execution_time()

            # We print the execution time.
            print(Fore.MAGENTA + Style.BRIGHT + "\nExecution time: " + formatted_time)

            percentage_output = (
                PyFunceble.OUTPUT_DIRECTORY
                + PyFunceble.OUTPUTS.parent_directory
                + PyFunceble.OUTPUTS.logs.directories.parent
                + PyFunceble.OUTPUTS.logs.directories.percentage
                + PyFunceble.OUTPUTS.logs.filenames.percentage
            )

            file_instance = PyFunceble.helpers.File(percentage_output)

            if PyFunceble.CONFIGURATION.show_percentage and file_instance.exists():
                file_instance.write(
                    "\nExecution time: {0}".format(formatted_time), overwrite=False
                )

        self.save(last=last)

    @classmethod
    def authorization(cls):
        """
        Checks the execution authorization.

        :return: The authorization status.
        :rtype: bool
        """

        if PyFunceble.CONFIGURATION.show_execution_time or PyFunceble.CONFIGURATION.ci:
            return True

        return False

    def save(self, last=False):
        """
        Saves the current time to our backup file.

        :param bool last:
            Tell us if we are at the very end of the file testing.
        """

        if (
            self.authorized
            and PyFunceble.CONFIGURATION.logs
            and "file_to_test" in PyFunceble.INTERN
            and PyFunceble.INTERN["file_to_test"]
        ):
            # * We are authorized to work.
            # and
            # * The generation of logs is activated.
            # and
            # * We are not testing as an imported module.

            # We set the location of the file we are working with.
            self.file = (
                PyFunceble.OUTPUT_DIRECTORY
                + PyFunceble.OUTPUTS.parent_directory
                + PyFunceble.OUTPUTS.logs.directories.parent
                + PyFunceble.OUTPUTS.logs.filenames.execution_time
            )

            if PyFunceble.helpers.File(self.file).exists():
                # The file we are working with exist.

                # We get its content so we can directly work with it.
                content = PyFunceble.helpers.Dict().from_json_file(self.file)
            else:
                # The file we are working with does not exist.

                # We generate a dummy content.
                content = {}

            if self.action == "start":
                # The action is equal to `start`.

                if "final_total" in content and content["final_total"]:
                    # The final total index exist.

                    # We delete it.
                    del content["final_total"]

                if "data" in content:
                    # The data index exist.

                    # We append the current start time inside it at
                    # a new sublist.
                    content["data"].append([PyFunceble.INTERN["start"]])
                else:
                    # The data index does not exist.

                    # We create the index along with the current start time.
                    content["data"] = [[PyFunceble.INTERN["start"]]]
            elif self.action == "stop":
                # The action is equal to `stop`.

                try:
                    # We try to work with the data index.

                    # We append the end time at the end of the last element
                    # of data.
                    #
                    # Note: It is at the end because we should have as first
                    # the star time.
                    content["data"][-1].append(PyFunceble.INTERN["end"])

                    # We get the start time.
                    start = content["data"][0][0]
                    # We get the end time.
                    end = content["data"][-1][-1]

                    # We calculate the execution time of the test.
                    content["current_total"] = self.format_execution_time(start, end)

                    if last:
                        # We are at the very end of the file testing.

                        # We initiate the global execution time.
                        content["final_total"] = content["current_total"]

                        # We inform the user about the global execution time.
                        print(
                            Fore.MAGENTA
                            + Style.BRIGHT
                            + "Global execution time: "
                            + content["final_total"]
                        )
                except KeyError:  # pragma: no cover
                    # It is not possible to work with the data index because
                    # it does not exist.

                    # We ignore the problem.
                    pass

            # We try to save the whole data at its final location.
            PyFunceble.helpers.Dict(content).to_json_file(self.file)

    @classmethod
    def set_starting_time(cls):  # pragma: no cover
        """
        Sets the starting time.
        """

        # We set the starting time as the current time.
        PyFunceble.INTERN["start"] = datetime.utcnow().timestamp()

        PyFunceble.LOGGER.debug(
            f'Starting time: {PyFunceble.INTERN["start"]} '
            f'| {datetime.fromtimestamp(PyFunceble.INTERN["start"])}'
        )

    @classmethod
    def set_stoping_time(cls):  # pragma: no cover
        """
        Sets the ending time.
        """

        # We set the ending time as the current time.
        PyFunceble.INTERN["end"] = datetime.utcnow().timestamp()

        PyFunceble.LOGGER.debug(
            f'Stoping time: {PyFunceble.INTERN["start"]} '
            f'| { datetime.fromtimestamp(PyFunceble.INTERN["end"])}'
        )

    @classmethod
    def calculate(cls, start=None, end=None):
        """
        Calculates the difference between starting and ending time.

        :param start: A starting time.
        :type start: int, str

        :param stop: A ending time.
        :type stop: int, str

        :return:
            A dict with following as index.

                * :code:`days`
                * :code:`hours`
                * :code:`minutes`
                * :code:`seconds`

            as index.
        :rtype: dict
        """

        if start and end:
            # The start and end time is explicitly given.

            # We get the difference between the ending and the starting time.
            time_difference = float(end) - float(start)
        else:
            # The start and end time is not explicitly given.

            # We get the difference between the ending and the starting time.
            time_difference = float(PyFunceble.INTERN["end"]) - float(
                PyFunceble.INTERN["start"]
            )

        # We initiate an OrderedDict.
        # Indeed, we use an ordered dict because we want the structuration and the
        # order to stay always the same.
        # As a dictionnary is always unordered, we can use it. Otherwise the time will
        # not be shown correctly.
        data = OrderedDict()

        # We calculate and append the day to our data.
        data["days"] = str(int(time_difference // (24 * 60 * 60))).zfill(2)

        # We calculate and append the hours to our data.
        data["hours"] = str(int((time_difference // (60 * 60)) % 24)).zfill(2)

        # We calculate and append the minutes to our data.
        data["minutes"] = str(int((time_difference % 3600) // 60)).zfill(2)

        # We calculate and append the minutes to our data.
        data["seconds"] = str(round(time_difference % 60, 6)).zfill(2)

        # We finaly return our data.
        return data

    def format_execution_time(self, start=None, end=None):
        """
        Formats the calculated time into a human readable format.

        :param start: A starting time.
        :type start: int, str

        :param stop: A ending time.
        :type stop: int, str

        :return: A human readable date.
        :rtype: str
        """

        # We return the formatted execution time.
        return ":".join(list(self.calculate(start, end).values()))



# pylint:disable=line-too-long
"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Tests of PyFunceble.extractor.expiration_date

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
# pylint: enable=line-too-long

from random import choice
from unittest import TestCase
from unittest import main as launch_tests

from PyFunceble.extractor import ExpirationDate


class TestExpirationDate(TestCase):
    """
    Tests of PyFunceble.engine.expiration_date.
    """

    def setUp(self):
        """
        Setups everything needed for the tests.
        """

        self.dates_samples = [
            "02-jan-2017",
            "02.01.2017",
            "02/01/2017",
            "2017-01-02",
            "2017.01.02",
            "2017/01/02",
            "2017.01.02 15:00:00",
            "20170102 15:00:00",
            "2017-01-02 15:00:00",
            "02.01.2017 15:00:00",
            "02-Jan-2017 15:00:00 UTC",
            "2017/01/02 01:00:00 (+0900)",
            "2017/01/02 01:00:00",
            "Mon Jan 02 15:00:00 GMT 2017",
            "Mon Jan 02 2017",
            "2017-01-02T15:00:00",
            "2017-01-02T15:00:00Z",
            "2017-01-02T15:00:00+0200",
            "2017-01-02T15:00:00+0200.622265+03:00",
            "2017-01-02T15:00:00+0200.622265",
            "2017-01-02T23:59:59.0Z",
            "02-01-2017",
            "2017. 01. 02.",
            "2017-01-02T00:00:00+13:00",
            "20170102",
            "02-Jan-2017",
            "02.1.2017",
            "02 Jan 2017",
            "02-January-2017",
            "2017-Jan-02.",
            "Mon Jan 02 15:00:00 2017",
            "January 02 2017-Jan-02",
            "2.1.2017",
            "20170102000000",
            "January  2 2017",
            "2nd January 2017",
        ]

        self.expiration_date_markers = [
            "expire: ",
            "expire on: ",
            "Expiry Date: ",
            "free-date ",
            "expires: ",
            "Expiration date: ",
            "Expiry date: ",
            "Expire Date: ",
            "renewal date: ",
            "Expires: ",
            "validity: ",
            "Expiration Date             : ",
            "Expiry : ",
            "expires at: ",
            "domain_datebilleduntil: ",
            "Data de expiração / Expiration Date (dd/mm/yyyy): ",
            "Fecha de expiración (Expiration date): ",
            "[Expires on] ",
            "status:      OK-UNTIL ",
            "renewal: ",
            "expires............: ",
            "expire-date: ",
            "Exp date: ",
            "Valid-date ",
            "Expires On: ",
            "Fecha de vencimiento: ",
            "Expiration:......... ",
            "Fecha de Vencimiento: ",
            "Registry Expiry Date: ",
            "Expires on..............: ",
            "Expiration Time: ",
            "Expiration Date: ",
            "Expired: ",
            "Date d'expiration: ",
            "expiration date: ",
        ]

        self.expiration_dates = []

        for marker in self.expiration_date_markers:
            self.expiration_dates.append(f"{marker} {choice(self.dates_samples)}")

    def test_extractor(self):
        """
        Tests extractor.
        """

        expected = "02-jan-2017"

        for date in self.expiration_dates:
            actual = ExpirationDate(date).get_extracted()

            self.assertEqual(expected, actual, date)

        actual = (
            ExpirationDate(
                f"Record expires on {choice(self.dates_samples)} (YYYY-MM-DD)"
            ).get_extracted(),
        )

        expected = None
        actual = ExpirationDate("").get_extracted()

        self.assertEqual(expected, actual)


if __name__ == "__main__":
    launch_tests()



# pylint:disable=line-too-long
"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides our exceptions.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""


class PyFuncebleException(Exception):
    """
    Describes our own exceptions.
    """


class PyFuncebleExternalException(PyFuncebleException):
    """
    Describes an exception which is caused by an external input.
    """


class PyFuncebleInternalException(PyFuncebleException):
    """
    Describes an exception which is caused by our own logic.
    """


class WrongParameterType(PyFuncebleInternalException):
    """
    Describes a wrong parameter type.
    """


class WrongParameterValue(PyFuncebleInternalException):
    """
    Describes a wrong parameter value.
    """


class NoInternetConnection(PyFuncebleExternalException):
    """
    Describes a missing connection.
    """


class UnableToDownload(PyFuncebleExternalException):
    """
    Describes a failing download.
    """


class ConfigurationFileNotFound(PyFuncebleExternalException):
    """
    Describes a missing configuration file.
    """


class GitHubTokenNotFound(PyFuncebleExternalException):
    """
    Describes a missing GitHub token.
    """


class GitLabTokenNotFound(PyFuncebleExternalException):
    """
    Describes a missing GitLab token.
    """


class GitEmailNotFound(PyFuncebleExternalException):
    """
    Describes a missing Git Email.
    """


class GitNameNotFound(PyFuncebleExternalException):
    """
    Describes a missing Git Name.
    """


class PleaseUpdatePyFunceble(PyFuncebleInternalException):
    """
    Describes the impossiblity to continue with an older version.
    """


class NoConversionMade(PyFuncebleInternalException):
    """
    Describes the fact that a conversion was expected but none
    was made.
    """


class NoExtractionMade(PyFuncebleInternalException):
    """
    Describes the fact that an extraction was expected but none
    was made.
    """


class UnknownSubject(PyFuncebleInternalException):
    """
    Describes the fact that an unknown subject is inputed.
    """


class NoDownloadDestinationGiven(PyFuncebleInternalException):
    """
    Describes the fact that the download destination was not declared.
    """


class NoDownloadLinkGiven(PyFuncebleInternalException):
    """
    Describes the fact that no download link was declared.
    """


class UserAgentNotFound(PyFuncebleInternalException):
    """
    Describes the fact that we could not find a user
    agent to work with.
    """


class UserAgentBrowserNotFound(PyFuncebleInternalException):
    """
    Describes the fact that we could not find a valid
    browser to work with.
    """


class UserAgentPlatformNotFound(PyFuncebleInternalException):
    """
    Describes the fact that we could not find a valid
    browser to work with.
    """



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides our hashes and position tracker.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from datetime import datetime

import PyFunceble


class HashesTracker:
    """
    Provides our hashes and position tracker.
    """

    tracker = dict()

    def __init__(self, filename):
        self.filename = filename

        self.tracker[filename] = dict()

        if self.authorized:
            self.hashes_file = (
                PyFunceble.CONFIG_DIRECTORY
                + PyFunceble.abstracts.Infrastructure.HASHES_FILENAME
            )

            PyFunceble.LOGGER.debug(f"Hashes file: {self.hashes_file}")

            self.load()
            self.__set_current_hash()
            self.__reset_if_changed()

    @property
    def authorized(self):
        """
        Provides the execution authorization.
        """

        return PyFunceble.CONFIGURATION.auto_continue

    def load(self):
        """
        Loads the hash file content.
        """

        file_destination = PyFunceble.helpers.File(self.hashes_file)

        if self.authorized and file_destination.exists():
            self.tracker = PyFunceble.helpers.Dict.from_json_file(file_destination.path)

            if self.filename not in self.tracker:
                self.tracker[self.filename] = dict()

            PyFunceble.LOGGER.info(f"Loaded {self.hashes_file!r} in memory.")

    def save(self):
        """
        Saves the current tracker into its file.
        """

        if self.authorized:
            PyFunceble.helpers.Dict(self.tracker).to_json_file(self.hashes_file)

            PyFunceble.LOGGER.info(f"Saved tracked into {self.hashes_file!r}.")

    def __set_current_hash(self):
        """
        Provides the hash of the current file.

        :rtype: str
        """

        if self.authorized:
            hash_datetime = datetime.utcnow().isoformat()

            hashes = PyFunceble.helpers.Hash().file(self.filename)

            self.tracker[self.filename]["latest"] = {
                "datetime_iso": hash_datetime,
                "hash": hashes,
            }

            if "base" not in self.tracker[self.filename]:
                self.tracker[self.filename]["base"] = {
                    "datetime_iso": hash_datetime,
                    "hash": hashes,
                }

            self.save()

            return hashes
        return None

    def hash_changed(self):
        """
        Checks if the current hash is different than our base one.

        :rtype: bool
        """

        if self.authorized:
            return (
                self.tracker[self.filename]["base"]["hash"]
                != self.tracker[self.filename]["latest"]["hash"]
            )
        return False

    def add_position(self, position):
        """
        Adds the given position to the previous one.
        """

        if self.authorized:
            try:
                self.tracker[self.filename]["position"] += position
            except KeyError:
                self.tracker[self.filename]["position"] = position

            self.save()

    def set_position(self, position):
        """
        Sets the position.
        """

        if self.authorized:
            self.tracker[self.filename]["position"] = position

            self.save()

    def get_position(self):
        """
        Provides the saved position.
        """

        try:
            return self.tracker[self.filename]["position"]
        except KeyError:
            return 0

    def reset_position(self):
        """
        Resets the position.
        """

        if self.authorized:
            self.set_position(0)

    def __reset_if_changed(self):
        """
        Resets the hashes.
        """

        if self.authorized and self.hash_changed():
            self.tracker[self.filename]["base"] = self.tracker[self.filename][
                "latest"
            ].copy()

            self.reset_position()

            self.save()



# pylint:disable=line-too-long
"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Tests of PyFunceble.converters.month

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
# pylint: enable=line-too-long
from unittest import TestCase
from unittest import main as launch_tests

import PyFunceble
from PyFunceble.converter.month import Month


class TestMonth(TestCase):
    """
    Tests of PyFunceble.converters.month
    """

    def test_wrong_type(self):
        """
        Tests for the case that the wrong type is given.
        """

        self.assertRaises(
            PyFunceble.exceptions.WrongParameterType,
            lambda: Month(["January"]).get_converted(),
        )

        self.assertRaises(
            PyFunceble.exceptions.WrongParameterType, lambda: Month(1).get_converted()
        )

    def test_january(self):
        """
        Tests for the case that we given a representation of the January month.
        """

        given = ["jan", "Jan", "January", "01", "1", "Jan.", "jan."]
        expected = "jan"

        for data in given:
            actual = Month(data).get_converted()

            self.assertEqual(expected, actual, data)

    def test_february(self):
        """
        Tests for the case that we given a representation of the February month.
        """

        given = ["feb", "Feb", "February", "02", "2", "Feb.", "feb."]
        expected = "feb"

        for data in given:
            actual = Month(data).get_converted()

            self.assertEqual(expected, actual, data)

    def test_march(self):
        """
        Tests for the case that we given a representation of the March month.
        """

        given = ["mar", "Mar", "March", "03", "3", "Mar.", "mar."]
        expected = "mar"

        for data in given:
            actual = Month(data).get_converted()

            self.assertEqual(expected, actual, data)

    def test_april(self):
        """
        Tests for the case that we given a representation of the April month.
        """

        given = ["apr", "Apr", "April", "04", "4", "Apr.", "apr."]
        expected = "apr"

        for data in given:
            actual = Month(data).get_converted()

            self.assertEqual(expected, actual, data)

    def test_may(self):
        """
        Tests for the case that we given a representation of the May month.
        """

        given = ["may", "May", "05", "5"]
        expected = "may"

        for data in given:
            actual = Month(data).get_converted()

            self.assertEqual(expected, actual, data)

    def test_june(self):
        """
        Tests for the case that we given a representation of the June month.
        """

        given = ["jun", "Jun", "June", "06", "6", "Jun.", "jun."]
        expected = "jun"

        for data in given:
            actual = Month(data).get_converted()

            self.assertEqual(expected, actual, data)

    def test_july(self):
        """
        Tests for the case that we given a representation of the July month.
        """

        given = ["jul", "Jul", "July", "07", "7", "Jul.", "jul."]
        expected = "jul"

        for data in given:
            actual = Month(data).get_converted()

            self.assertEqual(expected, actual, data)

    def test_august(self):
        """
        Tests for the case that we given a representation of the August month.
        """

        given = ["aug", "Aug", "August", "08", "8", "Aug.", "aug."]
        expected = "aug"

        for data in given:
            actual = Month(data).get_converted()

            self.assertEqual(expected, actual, data)

    def test_september(self):
        """
        Tests for the case that we given a representation of the September month.
        """

        given = [
            "sep",
            "Sep",
            "sept",
            "Sept",
            "September",
            "09",
            "9",
            "Sept.",
            "sept.",
            "Sep.",
            "sep.",
        ]
        expected = "sep"

        for data in given:
            actual = Month(data).get_converted()

            self.assertEqual(expected, actual, data)

    def test_october(self):
        """
        Tests for the case that we given a representation of the October month.
        """

        given = ["oct", "Oct", "October", "10", "Oct.", "oct."]
        expected = "oct"

        for data in given:
            actual = Month(data).get_converted()

            self.assertEqual(expected, actual, data)

    def test_november(self):
        """
        Tests for the case that we given a representation of the November month.
        """

        given = ["nov", "Nov", "November", "11", "Nov.", "nov."]
        expected = "nov"

        for data in given:
            actual = Month(data).get_converted()

            self.assertEqual(expected, actual, data)

    def test_december(self):
        """
        Tests for the case that we given a representation of the December month.
        """

        given = ["dec", "Dec", "December", "12", "Dec.", "dec."]
        expected = "dec"

        for data in given:
            actual = Month(data).get_converted()

            self.assertEqual(expected, actual, data)


if __name__ == "__main__":
    launch_tests()



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the directory helpers.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from os import getcwd, makedirs, path
from os import sep as directory_separator
from shutil import rmtree


class Directory:
    """
    Simplify the directories manipulation.

    :param str dir_path the path to work with.
    """

    def __init__(self, dir_path=None):
        self.path = dir_path

    @classmethod
    def get_current(cls, with_end_sep=False):
        """
        Returns the current directory path.

        :param bool with_end_sep:
            Add a directory separator at the end.
        """

        if with_end_sep:
            return getcwd() + directory_separator
        return getcwd()  # pragma: no cover

    def fix_path(self, dir_path=None, splited_path=None):
        """
        Fix the path of the given path.

        .. note::
            We consider a path as fixed if it ends with the right
            directory separator.
        """

        if not dir_path:
            dir_path = self.path

        if not splited_path:
            if dir_path:
                # The parsed directory is not empty or equal to None.

                # We initate a variable which will save the splited path.
                split_path = []

                if "/" in dir_path:
                    # We split the separator.
                    split_path = dir_path.split("/")
                elif "\\" in dir_path:
                    # We split the separator.
                    split_path = dir_path.split("\\")
                else:
                    split_path = [dir_path]
                # We join the splited element with the directory separator as glue.
                return self.fix_path(splited_path=[x for x in split_path if x])

            # We return the directory.
            return dir_path
        return directory_separator.join(splited_path) + directory_separator

    def exists(self, dir_path=None):
        """
        Checks if the given directory exists.
        """

        if not dir_path:
            dir_path = self.path

        return path.isdir(dir_path)

    def create(self, dir_path=None):  # pragma: no cover
        """
        Creates the given directory path.

        :return: The output of :code:`self.exists` after the directory creation.
        :rtype: bool
        """

        if not dir_path:
            dir_path = self.path

        try:
            if dir_path and not self.exists(dir_path=dir_path):
                makedirs(dir_path)
        except FileExistsError:
            pass

        return self.exists(dir_path=dir_path)

    def delete(self, dir_path=None):  # pragma: no cover
        """
        Deletes the given directory path.

        :return: :code:`not self.exists` after the directory deletion.
        :rtypt: bool
        """

        if not dir_path:
            dir_path = self.path

        if self.exists(dir_path=dir_path):
            rmtree(dir_path)

        return not self.exists(dir_path=dir_path)



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the API core interface.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import PyFunceble

from .cli import CLICore


class APICore:
    """
    Provides the logic and interface for the tests from the API.

    :param str subject: The element we are testing.
    :param bool complete:
        Activate the return of a dictionnary with significant - if not all -
        data about the test.

    :param dict configuration:
        The custom configuration to load.

        .. note::
            This will let you overwrite any of the previously given
            configuration index.
    """

    # The subject we are working with.
    subject = None
    # Tell us if we have to return all possible data.
    complete = False
    # Saves the configuration.
    configuration = None

    def __init__(
        self,
        subject,
        complete=False,
        configuration=None,
        db_file_name="api_call",
        is_parent=True,
    ):
        # We share the subject.
        self.subject = subject

        # We share the complete option.
        self.complete = complete

        # We share the given configuration.
        self.configuration = configuration

        # We share the db file name.
        self.db_file_name = db_file_name

        # We load the global configuration
        # if it was not alreay done.
        PyFunceble.load_config(
            generate_directory_structure=False, custom=self.configuration
        )

        if (
            "api_config_loaded" not in PyFunceble.INTERN
            or self.configuration != PyFunceble.INTERN["api_config_loaded"]
        ):
            # We update the configuration with the given
            # configuration.
            preset = PyFunceble.cconfig.Preset()
            preset.init_all()
            preset.api()

            PyFunceble.INTERN["api_config_loaded"] = (
                self.configuration.copy()
                if isinstance(self.configuration, dict)
                else dict()
            )

        # We create an instance of the whois database.
        self.whois_db = PyFunceble.database.Whois(parent_process=is_parent)

        # We create an instance of the inactive database.
        self.inactive_db = PyFunceble.database.Inactive(
            db_file_name, parent_process=is_parent
        )

    def __inactive_database_management(self, subject, status):
        """
        Given the subject and status, we add or remove the subject
        from the inactive database.
        """

        if self.inactive_db.authorized:
            # We are authorized to operate with the
            # inactive database.s

            if status.lower() in PyFunceble.STATUS.list.up:
                # The status is in the list of UP status.

                # We remove it from the database.
                self.inactive_db.remove(subject)
            else:
                # The status is not in the list of UP status.

                # We add it into the database.
                self.inactive_db.add(subject, status)

    def reputation(self, subject_type):
        """
        Make a reputation check.

        :param str subject_type:
            Should be one of the following.

            - :code:`domain`

            - :code:`url`
        """

        if isinstance(self.subject, list):
            # The given subject is a list of subjects.

            # We initiate a variable which save our result.
            result = {}

            for subject in self.subject:
                # We loop through the list of subject.

                result[subject] = APICore(
                    subject, complete=self.complete, configuration=self.configuration
                ).reputation(subject_type)

            # We return our local result.
            return result

        if "url" in subject_type:
            data = PyFunceble.status.UrlReputation(
                self.subject, whois_db=self.whois_db, inactive_db=self.inactive_db
            ).get()
        else:
            data = PyFunceble.status.DomainAndIPReputation(
                self.subject, whois_db=self.whois_db, inactive_db=self.inactive_db
            ).get()

        self.__inactive_database_management(self.subject, data["status"])
        CLICore.save_into_database(data, self.db_file_name)

        if self.complete:
            # The user want a copy of the compelte data.

            # We return them
            return data

        # We only return the status.
        return data["status"]

    def availability(self, subject_type):
        """
        Make an availability check.

        :param str subject_type:
            Should be one of the following.

            - :code:`domain`

            - :code:`url`
        """

        if "url" in subject_type:
            return self.url()
        return self.domain_and_ip()

    def syntax(self, subject_type):
        """
        Make a syntax check.

        :param str subject_type:
            Should be one of the following.

            - :code:`domain`

            - :code:`url`
        """

        if isinstance(self.subject, list):
            # The given subject is a list of subjects.

            # We initiate a variable which save our result.
            result = {}

            for subject in self.subject:
                # We loop through the list of subject.

                result[subject] = APICore(
                    subject, complete=self.complete, configuration=self.configuration
                ).syntax(subject_type)

            # We return our local result.
            return result

        if "url" in subject_type:
            data = PyFunceble.status.UrlSyntax(
                self.subject, whois_db=self.whois_db, inactive_db=self.inactive_db
            ).get()
        else:
            data = PyFunceble.status.DomainAndIpSyntax(
                self.subject, whois_db=self.whois_db, inactive_db=self.inactive_db
            ).get()

        self.__inactive_database_management(self.subject, data["status"])
        CLICore.save_into_database(data, self.db_file_name)

        if self.complete:
            # The user want a copy of the compelte data.

            # We return them
            return data

        # We only return the status.
        return data["status"]

    def domain_and_ip(self):
        """
        Run a domain/IP availability check over the given subject.
        """

        if isinstance(self.subject, list):
            # The given subject is a list of subjects.

            # We initiate a variable which save our result.
            result = {}

            for subject in self.subject:
                # We loop through the list of subject.

                result[subject] = APICore(
                    subject, complete=self.complete, configuration=self.configuration
                ).domain_and_ip()

            # We return our local result.
            return result

        # We get the status of the given subject.
        data = PyFunceble.status.DomainAndIpAvailability(
            self.subject, whois_db=self.whois_db, inactive_db=self.inactive_db
        ).get()

        self.__inactive_database_management(self.subject, data["status"])
        CLICore.save_into_database(data, self.db_file_name)

        if self.complete:
            # The user want a copy of the compelte data.

            # We return them
            return data

        # We only return the status.
        return data["status"]

    def domain_syntax(self):
        """
        Run a domain syntax check over the given subject.
        """

        if isinstance(self.subject, list):
            # The given subject is a list of subject.

            # We return the validity of each subjects.
            return {
                subject: APICore(
                    subject, complete=self.complete, configuration=self.configuration
                ).domain_syntax()
                for subject in self.subject
            }

        # We return the validity of the the given subject.
        return PyFunceble.Check(self.subject).is_domain()

    def subdomain_syntax(self):
        """
        Run a subdomain syntax check over the given subject.
        """

        if isinstance(self.subject, list):
            # The given subjet is a list of subject.

            # We return the validity of each subjects.
            return {
                subject: APICore(
                    subject, complete=self.complete, configuration=self.configuration
                ).subdomain_syntax()
                for subject in self.subject
            }

        # We return the validity of the given subject.
        return PyFunceble.Check(self.subject).is_subdomain()

    def ipv4_syntax(self):
        """
        Run an IPv4 syntax check over the given subject.
        """

        if isinstance(self.subject, list):
            # The given subjet is a list of subject.

            # We return the validity of each subjects.
            return {
                subject: APICore(
                    subject, complete=self.complete, configuration=self.configuration
                ).ipv4_syntax()
                for subject in self.subject
            }

        # We return the validity of the given subject.
        return PyFunceble.Check(self.subject).is_ipv4()

    def ipv6_syntax(self):
        """
        Run an IPv6 syntax check over the given subject.
        """

        if isinstance(self.subject, list):
            # The given subject is a list of subject.

            # We return the validity of each subjects.
            return {
                subject: APICore(
                    subject, complete=self.complete, configuration=self.configuration
                ).ipv6_syntax()
                for subject in self.subject
            }

        # We return the validity of the given subject.
        return PyFunceble.Check(self.subject).is_ipv6()

    def ip_syntax(self):
        """
        Run an IP syntax check over the given subject.
        """

        if isinstance(self.subject, list):
            # The given subject is a list of subject.

            # We return the validity of each subjects.
            return {
                subject: APICore(
                    subject, complete=self.complete, configuration=self.configuration
                ).ip_syntax()
                for subject in self.subject
            }

        # We return the validity of the given subject.
        return PyFunceble.Check(self.subject).is_ip()

    def ipv4_range_syntax(self):
        """
        Run an IPv4 range syntax check over the given subject.
        """

        if isinstance(self.subject, list):
            # The given subjet is a list of subject.

            # We return the validity of each subjects.
            return {
                subject: APICore(
                    subject, complete=self.complete, configuration=self.configuration
                ).ipv4_range_syntax()
                for subject in self.subject
            }

        # We return the validity of the given subject.
        return PyFunceble.Check(self.subject).is_ipv4_range()

    def ipv6_range_syntax(self):
        """
        Run an IPv6 range syntax check over the given subject.
        """

        if isinstance(self.subject, list):
            # The given subjet is a list of subject.

            # We return the validity of each subjects.
            return {
                subject: APICore(
                    subject, complete=self.complete, configuration=self.configuration
                ).ipv6_range_syntax()
                for subject in self.subject
            }

        # We return the validity of the given subject.
        return PyFunceble.Check(self.subject).is_ipv6_range()

    def ip_range_syntax(self):
        """
        Run an IP range syntax check over the given subject.
        """

        if isinstance(self.subject, list):
            # The given subjet is a list of subject.

            # We return the validity of each subjects.
            return {
                subject: APICore(
                    subject, complete=self.complete, configuration=self.configuration
                ).ip_range_syntax()
                for subject in self.subject
            }

        # We return the validity of the given subject.
        return PyFunceble.Check(self.subject).is_ip_range()

    def url(self):
        """
        Run an URL availability check over the given subject.
        """

        if isinstance(self.subject, list):
            # The given subjet is a list of subject.

            # We initiate a local variable which will save
            # what we are going to return.
            result = {}

            for subject in self.subject:
                # We loop through the list of subjects.

                result[subject] = APICore(
                    subject, complete=self.complete, configuration=self.configuration
                ).url()

            # We return the result of each subjects.
            return result

        # We get the complete data about the status of the subject.
        data = PyFunceble.status.UrlAvailability(
            self.subject, whois_db=self.whois_db, inactive_db=self.inactive_db
        ).get()

        self.__inactive_database_management(self.subject, data["status"])
        CLICore.save_into_database(data, self.db_file_name)

        if self.complete:
            # The user want a complete copy of the data.

            # We return them.
            return data

        # We return the result of each subjects.
        return data["status"]

    def url_syntax(self):
        """
        Run an IPv4 syntax check over the given subject.
        """

        if isinstance(self.subject, list):
            # The given subjet is a list of subject.

            # We return the validity of each subjects.
            return {
                subject: APICore(
                    subject, complete=self.complete, configuration=self.configuration
                ).url_syntax()
                for subject in self.subject
            }

        # We return the validity of the subject.
        return PyFunceble.Check(self.subject).is_url()



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the lookup interfaces.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from .dns import DNSLookup as Dns
from .http_code import HTTPCode
from .iana import Iana
from .ipv4_reputation import IPv4Reputation
from .publicsuffix import PublicSuffix
from .referer import Referer
from .requests import Requests
from .whois import WhoisLookup as Whois



# pylint:disable=line-too-long,invalid-name,import-error
"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Tests of the PyFunceble.core.cli

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
# pylint: enable=line-too-long

import sys
from datetime import datetime
from unittest import main as launch_tests
from unittest.mock import Mock, patch

from colorama import Fore, Style
from colorama import init as init_colorama
from stdout_base import StdoutBase
from time_zone import TZ

import PyFunceble
from PyFunceble.core import CLI


class TestCLICore(StdoutBase):
    """
    Tests of the PyFunceble.core.cli.
    """

    # pylint: disable=unsubscriptable-object

    def setUp(self):
        """
        Setups everything needed for the tests.
        """

        init_colorama(autoreset=True)

        PyFunceble.load_config()

        StdoutBase.setUp(self)

        self.green_ascii = "\n".join(
            [
                f"{Fore.GREEN}{x}{Fore.RESET}"
                for x in PyFunceble.ASCII_PYFUNCEBLE.split("\n")
            ]
        )
        self.red_ascii = "\n".join(
            [
                f"{Fore.RED}{x}{Fore.RESET}"
                for x in PyFunceble.ASCII_PYFUNCEBLE.split("\n")
            ]
        )
        self.home_ascii = "\n".join(
            [
                f"{Fore.YELLOW}{x}{Fore.RESET}"
                for x in PyFunceble.ASCII_PYFUNCEBLE.split("\n")
            ]
        )

        self.cli_core = CLI()

    def test_get_coloration(self):
        """
        Tests the method which give use the right coloration from a given status.
        """

        colors = {
            f"{Fore.GREEN}{Style.BRIGHT}": [
                PyFunceble.STATUS.official.up,
                PyFunceble.STATUS.official.valid,
            ],
            f"{Fore.RED}{Style.BRIGHT}": [PyFunceble.STATUS.official.down],
        }

        expected = f"{Fore.CYAN}{Style.BRIGHT}"
        actual = self.cli_core.get_simple_coloration(PyFunceble.STATUS.official.invalid)

        self.assertEqual(expected, actual)

        for expected, statuses in colors.items():
            for status in statuses:
                actual = self.cli_core.get_simple_coloration(status)

                self.assertEqual(expected, actual)

    def test_colored_ascii_home(self):
        """
        Tests the method which colors the ASCII representation of PyFunceble
        for the case that we want the home representation.
        """

        self.cli_core.colorify_logo(home=True)

        expected = f"{self.home_ascii}\n"
        actual = sys.stdout.getvalue()

        self.assertEqual(expected, actual)

    def test_colored_ascii_error(self):
        """
        Tests the method which colors the ASCII representation of PyFunceble
        for the case that we want the error representation.
        """

        PyFunceble.INTERN["counter"]["percentage"]["down"] = 55

        self.cli_core.colorify_logo()

        expected = f"{self.red_ascii}\n"
        actual = sys.stdout.getvalue()

        self.assertEqual(expected, actual)

        PyFunceble.INTERN["counter"]["percentage"]["down"] = 0

    def test_colored_ascii_success(self):
        """
        Tests the method which colors the ASCII representation of PyFunceble
        for the case that we want the success representation.
        """

        PyFunceble.INTERN["counter"]["percentage"]["up"] = 55

        self.cli_core.colorify_logo()

        expected = f"{self.green_ascii}\n"
        actual = sys.stdout.getvalue()

        self.assertEqual(expected, actual)

        PyFunceble.INTERN["counter"]["percentage"]["up"] = 0

    def test_colored_ascii_success_syntax(self):
        """
        Tests the method which colors the ASCII representation of PyFunceble
        for the case that we want the success representation for the case
        that we are testing for syntax.
        """

        PyFunceble.INTERN["counter"]["percentage"]["valid"] = 55

        self.cli_core.colorify_logo()

        expected = f"{self.green_ascii}\n"
        actual = sys.stdout.getvalue()

        self.assertEqual(expected, actual)

        PyFunceble.INTERN["counter"]["percentage"]["valid"] = 0

    def test_nothing_to_test(self):
        """
        Tests the correctness of the desired message.
        """

        expected = f"{Fore.CYAN}{Style.BRIGHT}Nothing to test.\n"
        self.cli_core.print_nothing_to_test()
        actual = sys.stdout.getvalue()

        self.assertEqual(expected, actual)

    @patch("datetime.datetime")
    def test_stay_safe(self, datetime_patch):
        """
        Tests the correctness of the desired message.
        """

        self.tearDown()
        self.setUp()

        expected = f"""
{Fore.GREEN}{Style.BRIGHT}Thanks for using PyFunceble!
"""

        datetime_patch = Mock(wraps=datetime)
        datetime_patch.utcnow = Mock(
            return_value=datetime(1970, 1, 1, 1, 0, 2, 0, tzinfo=TZ("+", hours=1).get())
        )
        patch("PyFunceble.core.cli.datetime", new=datetime_patch).start()

        CLI().stay_safe()
        actual = sys.stdout.getvalue()

        self.assertEqual(expected, actual)

    @patch("datetime.datetime")
    def test_stay_safe_share_experiance(self, datetime_patch):
        """
        Tests the correctness of the desired message.
        """

        self.tearDown()
        self.setUp()

        expected = f"""
{Fore.GREEN}{Style.BRIGHT}Thanks for using PyFunceble!
{Fore.YELLOW}{Style.BRIGHT}Share your experience on {Fore.CYAN}Twitter{Fore.YELLOW} with {Fore.CYAN}#PyFunceble{Fore.YELLOW}!
{Fore.GREEN}{Style.BRIGHT}Have a feedback, an issue or an improvement idea?
{Fore.YELLOW}{Style.BRIGHT}Let us know on {Fore.CYAN}GitHub{Fore.YELLOW}!
"""

        datetime_patch = Mock(wraps=datetime)
        datetime_patch.utcnow = Mock(
            return_value=datetime(1970, 1, 1, 1, 0, 3, 0, tzinfo=TZ("+", hours=1).get())
        )
        patch("PyFunceble.core.cli.datetime", new=datetime_patch).start()

        CLI().stay_safe()
        actual = sys.stdout.getvalue()

        self.assertEqual(expected, actual)

    def test_log_sharing(self):
        """
        Tests the correctness of the desired message.
        """

        PyFunceble.CONFIGURATION.share_logs = True

        expected = f"""{Fore.GREEN}{Style.BRIGHT}You are sharing your logs!
{Fore.MAGENTA}{Style.BRIGHT}Please find more about it at https://pyfunceble.readthedocs.io/en/dev/logs-sharing.html !
"""
        self.cli_core.logs_sharing()
        actual = sys.stdout.getvalue()

        self.assertEqual(expected, actual)


if __name__ == "__main__":
    launch_tests()



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the WHOIS database interface.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from datetime import datetime

from sqlalchemy.orm.exc import NoResultFound

import PyFunceble
from PyFunceble.engine.database.loader import session
from PyFunceble.engine.database.schemas import WhoisRecord


class WhoisDB:
    """
    Provides the WHOIS database interface and logic.
    """

    database = {}

    database_file = None

    def __init__(self, parent_process=False):
        # Get the authorization.
        self.database_file = ""

        if PyFunceble.CONFIGURATION.db_type == "json":
            # We set the location of the database file.
            self.database_file = "{0}{1}".format(
                PyFunceble.CONFIG_DIRECTORY, PyFunceble.OUTPUTS.default_files.whois_db
            )

        self.parent = parent_process

        PyFunceble.LOGGER.debug(f"DB (File): {self.database_file}")

        # We load the configuration.
        self.load()

    def __contains__(self, index):
        if self.authorized:
            if PyFunceble.CONFIGURATION.db_type == "json":
                if index in self.database:
                    PyFunceble.LOGGER.info(f"{index} is present into the database.")
                    return True

                PyFunceble.LOGGER.info(f"{index} is not present into the database.")
                return False

            if PyFunceble.CONFIGURATION.db_type in ["mariadb", "mysql"]:
                with session.Session() as db_session:
                    try:
                        # pylint: disable=no-member
                        _ = (
                            db_session.query(WhoisRecord).filter(
                                WhoisRecord.subject == index
                            )
                        ).one()

                        PyFunceble.LOGGER.info(f"{index} is present into the database.")
                        return True
                    except NoResultFound:
                        PyFunceble.LOGGER.info(
                            f"{index} is not present into the database."
                        )

        return False  # pragma: no cover

    def __getitem__(self, index):
        if self.authorized:
            if PyFunceble.CONFIGURATION.db_type == "json":
                if index in self.database:
                    return self.database[index]
                return None

            if PyFunceble.CONFIGURATION.db_type in ["mariadb", "mysql"]:
                fetched = None

                with session.Session() as db_session:
                    try:
                        # pylint: disable=no-member
                        fetched = (
                            db_session.query(WhoisRecord).filter(
                                WhoisRecord.subject == index
                            )
                        ).one()

                        return {
                            "epoch": fetched.epoch,
                            "expiration_date": fetched.expiration_date,
                            "state": fetched.state,
                            "record": fetched.record,
                        }
                    except NoResultFound:
                        pass

        return None  # pragma: no cover

    def __setitem_json(self, index, value):
        actual_value = self[index]

        if not PyFunceble.CONFIGURATION.store_whois_record and "record" in value:
            del value["record"]

        if isinstance(actual_value, dict):
            if isinstance(value, dict):
                self.database[index].update(value)
            else:  # pragma: no cover
                self.database[index] = value
        elif isinstance(actual_value, list):
            if isinstance(value, list):  # pragma: no cover
                self.database[index].extend(value)
            else:  # pragma: no cover
                self.database[index].append(value)
        else:
            self.database[index] = value

        PyFunceble.LOGGER.info(
            f"Inserted {repr(value)} into the subset of {repr(index)}"
        )

    @classmethod
    def __setitem_mysql(cls, index, value):
        with session.Session() as db_session:
            try:
                # pylint: disable=no-member
                record = (
                    db_session.query(WhoisRecord)
                    .filter(WhoisRecord.subject == index)
                    .one()
                )
            except NoResultFound:
                record = WhoisRecord(subject=index,)

        for db_key, db_value in value.items():
            if not PyFunceble.CONFIGURATION.store_whois_record and db_key == "record":
                continue
            setattr(record, db_key, db_value)

        with session.Session() as db_session:
            # pylint: disable=no-member
            db_session.add(record)
            db_session.commit()

            PyFunceble.LOGGER.info(f"Inserted into the database: \n {value}")

    def __setitem__(self, index, value):
        if self.authorized:
            if PyFunceble.CONFIGURATION.db_type == "json":
                self.__setitem_json(index, value)
            elif PyFunceble.CONFIGURATION.db_type in ["mariadb", "mysql"]:
                self.__setitem_mysql(index, value)

    @property
    def authorized(self):
        """
        Provides the operation authorization.
        """

        return (
            not PyFunceble.CONFIGURATION.no_whois
            and PyFunceble.CONFIGURATION.whois_database
        )

    @classmethod
    def merge(cls, old):  # pragma: no cover
        """
        Merges the older version of the database into the new version.

        :param dict old: The old version of the database.


        :return: The database in the new format.
        :rtype: dict
        """

        # We initiate a local place to save our results.
        result = {}

        if PyFunceble.CONFIGURATION.db_type == "json":
            for index, data in old.items():
                # We loop through all indexes and data of the database.

                if isinstance(data, dict) and "epoch" in data:
                    # The epoch index is present into the currently
                    # read dataset.

                    # We create the copy of the dataset for our result.
                    result[index] = data

                    continue

                if isinstance(data, dict):
                    # The read data is a dict.

                    # We save the content of of the currently read dataset
                    # into the upstream index.
                    result.update(data)

        # We return the result.
        return result

    def load(self):
        """
        Loads the database file into the database.
        """

        if (
            self.authorized
            and PyFunceble.helpers.File(self.database_file).exists()
            and PyFunceble.CONFIGURATION.db_type == "json"
        ):
            # * We are authorized to operate.
            # and
            # * The database file exists.

            # We merge our current database into already initiated one.
            self.database.update(
                self.merge(PyFunceble.helpers.Dict().from_json_file(self.database_file))
            )

            PyFunceble.LOGGER.info(
                "Database content loaded in memory. (DATASET WONT BE LOGGED)"
            )

            # As changes can happen because of the merging, we directly saved
            # the loaded data.
            self.save()

    def save(self):
        """
        Saves the database into the database file.
        """

        if (
            self.authorized
            and self.parent
            and PyFunceble.CONFIGURATION.db_type == "json"
        ):
            # We are authorized to operate.

            # We save the current state of the datbase.
            PyFunceble.helpers.Dict(self.database).to_json_file(self.database_file)

            PyFunceble.LOGGER.info(f"Saved database into {repr(self.database_file)}.")

    def is_time_older(self, subject):
        """
        Checks if the expiration time of the given subject is
        older.

        :param str subject: The subject we are working with.

        .. note::
            Of course, we imply that the subject is in the database.
        """

        data = self[subject]

        return (
            self.authorized
            and data
            and "epoch" in data
            and datetime.fromtimestamp(float(data["epoch"])) < datetime.utcnow()
        )

    def get_expiration_date(self, subject):
        """
        Gets the expiration date of the given subject.

        :param str subject: The subject we are working with.

        :return: The expiration date from the database.
        :rtype: tuple
        """

        if self.authorized and self[subject] and not self.is_time_older(subject):
            # * We are authorized to work.
            # and
            # * The element we are testing is in the database.
            # and
            # * The expiration date is in the future.

            if "record" in self[subject]:
                whois_record = self[subject]["record"]
            else:
                whois_record = None

            try:
                # We return the expiration date.
                return self[subject]["expiration_date"], whois_record
            except KeyError:  # pragma: no cover
                pass

        # We return None, there is no data to work with.
        return None, None

    def add(self, subject, expiration_date, server, record=None):
        """
        Adds the given subject and expiration date to the database.

        :param str subject: The subject we are working with.
        :param str expiration_date: The extracted expiration date.
        :param str server: The server we got the record from.
        :param str record: The WHOIS record.
        """

        if self.authorized and expiration_date:
            # * We are authorized to operate.
            # and
            # * The expiration date is not empty nor None.

            # We initiate what we are going to save into the database
            data = {
                "epoch": datetime.strptime(expiration_date, "%d-%b-%Y").timestamp(),
                "expiration_date": expiration_date,
                "server": server,
            }

            if record:  # pragma: no cover
                data["record"] = str(record)
            else:
                data["record"] = str(None)

            if data["epoch"] < datetime.utcnow().timestamp():
                # We compare the epoch with the current time.

                # We set the state.
                data["state"] = "past"
            else:
                # We set the state.
                data["state"] = "future"

            # We save everything into the database.
            self[subject] = data

            # We save everything.
            self.save()



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the base of all our SQLAlchemy schemas.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from datetime import datetime

from inflection import underscore
from sqlalchemy import Column, DateTime, Integer
from sqlalchemy.ext.declarative import declarative_base, declared_attr


class CustomClass:
    """
    Provides the base of all our schemas.
    """

    @declared_attr
    def __tablename__(cls):  # pylint: disable=no-self-argument
        """
        Maps the class name with the table name.
        """

        # pylint: disable=no-member
        return "pyfunceble_" + underscore(cls.__name__).lower()

    id = Column(Integer, primary_key=True, nullable=False)
    created = Column(DateTime(), default=datetime.utcnow, nullable=False)
    modified = Column(DateTime(), default=datetime.utcnow, onupdate=datetime.utcnow)


DatabaseBase = declarative_base(cls=CustomClass)



# pylint:disable=line-too-long
"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Tests of PyFunceble.converters.file

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
# pylint: enable=line-too-long
from unittest import TestCase
from unittest import main as launch_tests

import PyFunceble
from PyFunceble.converter.file import File


class TestFileLineConverter(TestCase):
    """
    Tests of PyFunceble.converter.file
    """

    def setUp(self):
        """
        Setups everything that is needed for the tests.
        """

        PyFunceble.load_config(generate_directory_structure=False)

        self.domains = [
            "google.com",
            "twitter.com",
            "github.com",
            "facebook.com",
            "hello.world",
            "world.hello",
        ]

    def tests_simple_line(self):
        """
        Tests the case that we encouter a simple line.
        """

        for domain in self.domains:
            expected = domain
            actual = File(domain).get_converted()

            self.assertEqual(expected, actual)

    def tests_comment(self):
        """
        Tests the case that we encouter a commented line.
        """

        for domain in self.domains:
            expected = None

            data = f"# {domain}"
            actual = File(data).get_converted()

            self.assertEqual(expected, actual)

    def tests_ends_with_comment(self):
        """
        Tests the case that a line has a comment at the end of its line.
        """

        for domain in self.domains:
            expected = domain

            data = f"{domain} # hello world"
            actual = File(data).get_converted()

            self.assertEqual(expected, actual)

    def tests_with_prefix(self):
        """
        Tests the case that a line has a decorator.

        For example:
        ::

            127.0.0.1 google.com
        """

        for domain in self.domains:
            expected = domain

            data = f"0.0.0.0 {domain}"
            actual = File(data).get_converted()

            self.assertEqual(expected, actual)

        for domain in self.domains:
            expected = domain

            data = f"127.0.0.1 {domain}"
            actual = File(data).get_converted()

            self.assertEqual(expected, actual)

    def tests_with_prefix_and_multiple_subjects(self):
        """
        Tests the case that a line has a decorator and multiple subjects.

        For example:
        ::

            127.0.0.1 google.com example.com example.net
        """

        for domain in self.domains:
            expected = [domain, "example.com", "example.net"]

            data = f"0.0.0.0 {domain} example.com\t\texample.net"
            actual = File(data).get_converted()

            self.assertEqual(expected, actual)

        for domain in self.domains:
            expected = [domain, "example.com", "example.net"]

            data = f"127.0.0.1 {domain} example.com\t\texample.net"
            actual = File(data).get_converted()

            self.assertEqual(expected, actual)

    def tests_multiple_spaces(self):
        """
        Tests the case that we have multiple space as sparator between
        our domain end its prefix.
        """

        for domain in self.domains:
            expected = domain

            data = f"0.0.0.0                {domain}"
            actual = File(data).get_converted()

            self.assertEqual(expected, actual)

        for domain in self.domains:
            expected = domain

            data = f"127.0.0.1                {domain}"
            actual = File(data).get_converted()

            self.assertEqual(expected, actual)

    def tests_with_tabs(self):
        """
        Tests the case that we have a single tab as sparator between
        our domain end its prefix.
        """

        for domain in self.domains:
            expected = domain

            data = f"0.0.0.0\t{domain}"
            actual = File(data).get_converted()

            self.assertEqual(expected, actual)

        for domain in self.domains:
            expected = domain

            data = f"127.0.0.1\t{domain}"
            actual = File(data).get_converted()

            self.assertEqual(expected, actual)

    def tests_with_multiple_tabs(self):
        """
        Tests the case that we have multiple tabs as sparator between
        our domain end its prefix.
        """

        for domain in self.domains:
            expected = domain

            data = f"0.0.0.0\t\t\t\t\t\t\t\t\t\t{domain}"
            actual = File(data).get_converted()

            self.assertEqual(expected, actual)

        for domain in self.domains:
            expected = domain

            data = f"127.0.0.1\t\t\t\t\t\t\t\t\t\t\t{domain}"
            actual = File(data).get_converted()

            self.assertEqual(expected, actual)

    def tests_with_list_of_entries(self):
        """
        Tests the case that we give multiple entries
        """

        given = [f"0.0.0.0\t\t\t\t\t\t\t\t\t\t{domain}" for domain in self.domains]
        expected = self.domains
        actual = File(given).get_converted()

        self.assertEqual(expected, actual)


if __name__ == "__main__":
    launch_tests()



# pylint:disable=line-too-long
"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Tests of PyFunceble.engine.sort

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
# pylint: enable=line-too-long

from unittest import TestCase
from unittest import main as launch_tests

from PyFunceble.engine import Sort
from PyFunceble.helpers import List


class TestSort(TestCase):
    """
    Tests of PyFunceble.engine.sort
    """

    def setUp(self):
        """
        Setups everything needed for the tests.
        """

        self.data_list = [
            "adservice.google.com",
            "pagead2.googleadservices.com",
            "ade.googlesyndication.com",
            "1.gravatar.com",
            "chart.googleapis.com",
            "imasdk.googleapis.com",
            "google.com",
            "ad-creatives-public.commondatastorage.googleapis.com",
            "googleadservices.com",
            "www-google-analytics.l.google.com",
            "0.gravatar.com",
            "hello_world.google.com",
            "partner.googleadservices.com",
            "www.googleadservices.com",
            "s0-2mdn-net.l.google.com",
            "ssl-google-analytics.l.google.com",
            "google-analytics.com",
            "redirector.googlevideo.com",
            "www.google-analytics.com",
            "ssl.google-analytics.com",
            "pagead2.googlesyndication.com",
            "www.googletagmanager.com",
            "tpc.googlesyndication.com",
            "www.googletagservices.com",
            "hello",
        ]

        self.data_url_list = [
            "https://0.gravatar.com",
            "https://1.gravatar.com",
            "https://ad-creatives-public.commondatastorage.googleapis.com",
            "https://ade.googlesyndication.com",
            "https://adservice.google.com",
            "https://chart.googleapis.com",
            "https://google.com",
            "https://google-analytics.com",
            "https://googleadservices.com",
            "https://hello_world.google.com",
            "https://imasdk.googleapis.com",
            "https://pagead2.googleadservices.com",
            "https://pagead2.googlesyndication.com",
            "https://partner.googleadservices.com",
            "https://redirector.googlevideo.com",
            "https://s0-2mdn-net.l.google.com",
            "https://ssl-google-analytics.l.google.com",
            "https://ssl.google-analytics.com",
            "https://tpc.googlesyndication.com",
            "https://www-google-analytics.l.google.com",
            "https://www.google-analytics.com",
            "https://www.googleadservices.com",
            "https://www.googletagmanager.com",
            "https://www.googletagservices.com",
            "https://hello",
        ]

    def test_standard_numeric_sorting(self):
        """
        Tests the standard numeric sorting.
        """

        given = [
            "1.example.com",
            "2.example.com",
            "3.example.com",
            "11.example.com",
            "12.example.com",
            "10.example.com",
        ]

        given_url = [f"http://{x}" for x in given]

        expected = [
            "1.example.com",
            "2.example.com",
            "3.example.com",
            "10.example.com",
            "11.example.com",
            "12.example.com",
        ]

        expected_url = [f"http://{x}" for x in expected]

        actual = List(given).custom_format(Sort.standard)

        self.assertEqual(expected, actual)

        actual_url = List(given_url).custom_format(Sort.standard)

        self.assertEqual(expected_url, actual_url)

    def test_hierarchical_numeric_sorting(self):
        """
        Tests the hierarchical numeric sorting.
        """

        given = [
            "1.example.com",
            "2.example.com",
            "3.example.com",
            "11.example.com",
            "12.example.com",
            "10.example.com",
            "hello.1.example.com",
            "hello.2.example.com",
            "hello.3.example.com",
            "hello.11.example.com",
            "hello.12.example.com",
            "hello.10.example.com",
        ]

        expected = [
            "1.example.com",
            "hello.1.example.com",
            "2.example.com",
            "hello.2.example.com",
            "3.example.com",
            "hello.3.example.com",
            "10.example.com",
            "hello.10.example.com",
            "11.example.com",
            "hello.11.example.com",
            "12.example.com",
            "hello.12.example.com",
        ]

        actual = List(given).custom_format(Sort.hierarchical)

        self.assertEqual(expected, actual)

    def test_standard_sorting(self):
        """
        Tests standard sorting.
        """

        expected = [
            "0.gravatar.com",
            "1.gravatar.com",
            "ad-creatives-public.commondatastorage.googleapis.com",
            "ade.googlesyndication.com",
            "adservice.google.com",
            "chart.googleapis.com",
            "google.com",
            "googleadservices.com",
            "google-analytics.com",
            "hello",
            "hello_world.google.com",
            "imasdk.googleapis.com",
            "pagead2.googleadservices.com",
            "pagead2.googlesyndication.com",
            "partner.googleadservices.com",
            "redirector.googlevideo.com",
            "s0-2mdn-net.l.google.com",
            "ssl.google-analytics.com",
            "ssl-google-analytics.l.google.com",
            "tpc.googlesyndication.com",
            "www.googleadservices.com",
            "www.google-analytics.com",
            "www.googletagmanager.com",
            "www.googletagservices.com",
            "www-google-analytics.l.google.com",
        ]
        actual = List(self.data_list).custom_format(Sort.standard)

        self.assertEqual(expected, actual)

        expected_url = [f"https://{x}" for x in expected]
        actual_url = List(self.data_url_list).custom_format(Sort.standard)

        self.assertEqual(expected_url, actual_url)

    def test_hierarchical_sorting(self):
        """
        Tests hierarchical sorting.
        """

        expected = [
            "google.com",
            "adservice.google.com",
            "hello_world.google.com",
            "s0-2mdn-net.l.google.com",
            "ssl-google-analytics.l.google.com",
            "www-google-analytics.l.google.com",
            "googleadservices.com",
            "pagead2.googleadservices.com",
            "partner.googleadservices.com",
            "www.googleadservices.com",
            "google-analytics.com",
            "ssl.google-analytics.com",
            "www.google-analytics.com",
            "chart.googleapis.com",
            "ad-creatives-public.commondatastorage.googleapis.com",
            "imasdk.googleapis.com",
            "ade.googlesyndication.com",
            "pagead2.googlesyndication.com",
            "tpc.googlesyndication.com",
            "www.googletagmanager.com",
            "www.googletagservices.com",
            "redirector.googlevideo.com",
            "0.gravatar.com",
            "1.gravatar.com",
            "hello",
        ]

        actual = List(self.data_list).custom_format(Sort.hierarchical)

        self.assertEqual(expected, actual)

        expected_url = [f"https://{x}" for x in expected]

        actual_url = List(self.data_url_list).custom_format(Sort.hierarchical)

        self.assertEqual(expected_url, actual_url)


if __name__ == "__main__":
    launch_tests()



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the WHOIS lookup interface.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from random import choice
from socket import AF_INET, SOCK_STREAM
from socket import error as socket_error
from socket import socket
from socket import timeout as socket_timeout

import PyFunceble

from .referer import Referer


class WhoisLookup:
    """
    Implementation of the UNIX `whois` command.

    :param str subject: The subject we are working with.

    :param str server:
        The WHOIS server we are working with.

        .. note::
            If :code:`None` is given, we look for the best one.

    :param int timeout:
        The timeout to apply.

        .. warning::
            The timeout must be a modulo of :code:`3`.
    """

    # Set the port we are going to working with.
    universal_port = 43
    # Set the size of the buffer which extrating
    # the expiration date.
    buffer_size = 4096

    def __init__(self, subject, server=None, timeout=3):
        if subject:
            # The subject is not empty nor None.

            if isinstance(subject, str):
                # The subject is a str.

                # We share it.
                self.subject = subject
            else:
                # The subject is not a str.

                # We raise an exception.
                raise ValueError("expected {0}".format(type(str)))

        if server is not None:
            # The server is given.

            if isinstance(server, str):
                # The server is a str.

                resolved_server = PyFunceble.DNSLOOKUP.a_record(server)

                try:
                    # We share it.
                    self.server = choice(resolved_server)
                except (IndexError, TypeError):
                    self.server = None
            else:
                # The server is not a str.

                # We raise an exception.
                raise ValueError("`server` must be a string.")
        else:
            # The server is not given or is None.

            # We get the server.
            self.server = Referer(self.subject).get()[0]

        if timeout:
            # The timeout is given.

            if isinstance(timeout, (int, float)):
                # The timeout is an int

                # We share it.
                self.timeout = float(timeout)
            else:
                # The timeout is something we could not understand.

                # We eaise an exception.
                raise ValueError("`timeout` must be an integer or float.")

    def request(self):  # pragma: no cover
        """
        Perform the WHOIS request.
        """

        result = None

        if self.server and self.subject:
            # A server was found or is given.

            # We initiate a socket for the request.
            req = socket(AF_INET, SOCK_STREAM)

            # We set the timeout.
            req.settimeout(self.timeout)

            try:
                # We try to connect to the whois server at the port 43.
                req.connect((self.server, self.universal_port))
            except socket_error:
                return result

            # We send and encode the domain we want the information from.
            req.send("{}\r\n".format(self.subject).encode())

            # We initiate a bytes variable which will save the response from the server.
            response = bytes("", "utf-8")

            while True:
                # We loop infinitly

                try:
                    data = req.recv(self.buffer_size)
                except (ConnectionResetError, socket_timeout):
                    # We got an error.

                    # We close the connection.
                    req.close()

                    # And we return the result.
                    return result

                # Eerything goes right.

                # We append the data to the response.
                response += data

                if not data:
                    # The data is empty or equal to None.

                    # We close the connection.
                    req.close()

                    # And we break the loop.
                    break

            try:
                # We finally decode and return the response we got from the server.

                return response.decode()
            except UnicodeDecodeError:
                # We may get a decoding error.

                # We decode the response explicitly.
                # Note: Because we don't want to deal with other issue, we
                # decided to use `replace` in order to automatically replace
                # all non utf-8 encoded characters.
                return response.decode("utf-8", "replace")

        # The whois server is not given nor found.

        return result



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides our database session interface.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from .credential import Credential


class Session:
    """
    Provides our database session interface,
    """

    migration_effective = False

    def __init__(self):
        self.credentials = Credential()
        self.engine = create_engine(
            self.credentials.get_uri(), pool_pre_ping=True, pool_recycle=180
        )
        self.session = sessionmaker(autocommit=False, autoflush=False, bind=self.engine)

        self.current_session = None

    def __enter__(self):
        self.create_new_session()

        return self.current_session

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def query(self, *args, **kwargs):
        """
        Make a query.
        """

        return self.create_new_session().query(*args, **kwargs)

    def close(self):
        """
        Closes the session.
        """

        if self.current_session is not None:
            # pylint: disable=no-member
            self.current_session.close()

    def create_new_session(self):
        """
        Provides a new session.
        """

        if self.current_session is not None:
            self.current_session.close()
            self.current_session = None

        self.current_session = self.session()

        return self.current_session



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the list helpers

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""


class List:
    """
    Simplify the list manipulation.

    :param list main: The main list to work with.
    :param bool remove_empty: Process the deletion of empty strings.
    """

    def __init__(self, main, remove_empty=False):
        if remove_empty:
            self.main = [x for x in main if x is None or x]
        else:
            self.main = main

    def format(self):
        """
        Return a well formatted list. Basicaly, it's sort a list and remove duplicate.

        :return: A sorted, without duplicate, list.
        :rtype: list
        """

        try:
            return sorted(list(set(self.main)), key=str.lower)

        except TypeError:  # pragma: no cover
            return self.main

    def custom_format(self, key_method, reverse=False):
        """
        Return a well formatted list. With the key_method as a function/method to format
        the elements before sorting.

        :param key_method:
            A function or method to use to format the
            readed element before sorting.
        :type key_method: function|method

        :param bool reverse: Tell us if we have to reverse the list.

        :return: A sorted list.
        :rtype: list
        """

        try:
            return sorted(list(set(self.main)), key=key_method, reverse=reverse)
        except TypeError:  # pragma: no cover
            return self.main



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides everything related to the package and its version.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from PyFunceble import helpers


class Package:
    """
    Provides some packaging related abstractions.
    """

    NAME = "PyFunceble"
    """
    Sets the package name.

    :type: str
    """

    VERSION = "3.3.0.dev (Teal Blauwbok: Centipede)"
    """
    Sets the package version.

    :type: str
    """


class Version:
    """
    Provides a simple way to compare our own versions.
    """

    @classmethod
    def split_versions(cls, version, return_non_digits=False):
        """
        Converts the versions to a shorter one.

        :param str version: The version to split.

        :param bool return_non_digits:
            Activate the return of the non-digits parts of the splitted
            version.

        :return:
            A tuple. The first index is the digit part of the version,
            when the second one is the the non-digit part of the
            version.
        :rtype: tuple
        """

        # We split the version.
        splited_version = version.split(".")

        # We keep the digits only.
        digits = [x for x in splited_version if x.isdigit()]

        if not return_non_digits:
            # We do not have to return the non digits part of the version.

            # We return the digits part of the version.
            return digits

        # We have to return the non digit parts of the version.

        # We keep the non digits.
        non_digits = [x for x in splited_version if not x.isdigit()]

        # We return a tuple with first the digits part and finally the non digit parts.
        return digits, non_digits[0]

    @classmethod
    def literally_compare(cls, local, upstream):
        """
        Compares the given versions, literally.

        :param str local:
            The local version converted
            by :py:func:`~PyFunceble.abstracts.package.split_versions`.

        :param str upstream:
            The upstream version converted
            by :py:func:`~PyFunceble.abstracts.package.split_versions`.

        :return:
            - :code:`True`: local == upstream
            - :code:`False`: local != upstream
        :rtype: bool
        """

        return local == upstream

    @classmethod
    def compare(cls, upstream):
        """
        Compares the given versions with the local one.

        :param list local:
            The local version converted
            by :py:func:`~PyFunceble.abstracts.package.split_versions`.

        :param list upstream:
            The upstream version converted
            by :py:func:`~PyFunceble.abstracts.package.split_versions`.

        :return:
            - :code:`True`: local < upstream
            - :code:`None`: local == upstream
            - :code:`False`: local > upstream
        :rtype: bool, None
        """

        # We get the local version.
        local = cls.split_versions(Package.VERSION)
        # We get the upstream version
        upstream = cls.split_versions(upstream)

        # A version should be in format [1,2,3] which is actually the version `1.2.3`
        # So as we only have 3 elements in the versioning,
        # we initiate the following variable in order to get the status of each parts.
        status = [None, None, None]

        for index, version_number in enumerate(local):
            # We loop through the local version.

            if int(version_number) < int(upstream[index]):
                # The local version is less than the upstream version.

                # We initiate its status to True which means that we are in
                # an old version (for the current version part).
                status[index] = True
            elif int(version_number) > int(upstream[index]):
                # The local version is greater then the upstream version.

                # We initiate its status to False which means that we are in
                # a more recent version (for the current version part).
                status[index] = False
            else:
                # The local version is eqal to the upstream version.

                # We initiate its status to None which means that we are in
                # the same version (for the current version part).
                status[index] = None

            # Otherwise the status stay None which means that there is no change
            # between both local and upstream.

        # We consider that the version is the same.
        result = None

        for data in status:
            # We loop through the list of status.
            # The purpose of this loop is only to
            # get the first not None value.

            if result is None:
                # The result is None (no changes).
                # We set the currently read one as the result.
                result = data

        # We return the result.
        return result

    @classmethod
    def is_local_dev(cls):
        """
        Checks if the local version is the development version.
        """

        return "dev" in Package.VERSION

    @classmethod
    def is_local_cloned(cls):  # pragma: no cover
        """
        Checks if the local version is was downloaded
        per :code:`git clone`.
        """

        if not helpers.Directory(".git").exists():
            # The git directory does not exist.

            # We return False, the current version is not the cloned version.
            return False

        # We list the list of file which can be found only in a cloned version.
        list_of_file = [
            ".coveragerc",
            ".coveralls.yml",
            ".gitignore",
            ".PyFunceble_production.yaml",
            ".travis.yml",
            "CODE_OF_CONDUCT.rst",
            "CONTRIBUTING.rst",
            "dir_structure_production.json",
            "MANIFEST.in",
            "README.rst",
            "requirements.txt",
            "setup.py",
            "version.yaml",
        ]

        # We list the list of directory which can be found only in a cloned
        # version.
        list_of_dir = ["docs", "PyFunceble", "tests"]

        if not all([helpers.File(x).exists() for x in list_of_file]):
            return False

        # All required files exist in the current directory.

        if not all([helpers.Directory(x).exists() for x in list_of_dir]):
            return False

        # All required directories exist in the current directory.

        # We return True, the current version is a cloned version.
        return True



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the base of all our downloader.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import sys
from datetime import datetime, timedelta

from colorama import Fore, Style

import PyFunceble

from ..exceptions import UnableToDownload


class DownloaderBase:
    """
    Provides the downloader base.
    """

    # pylint: disable=no-member

    DOWNTIME_INDEX = None
    """
    Used to set/track the download time of the current file.

    :rtype: str
    """

    REDOWNLOAD_AFTER = 1
    """
    Used to set the redownload frequency in day.

    :rtype: int
    """

    def __init__(self):
        self.downtimes_file = PyFunceble.helpers.File(
            f"{PyFunceble.CONFIG_DIRECTORY}{PyFunceble.abstracts.Infrastructure.DOWN_FILENAME}"
        )

        self.all_downtimes = self.get_all_downtimes()

        if not hasattr(self, "destination"):
            raise PyFunceble.exceptions.NoDownloadDestinationGiven()

        if not hasattr(self, "download_link"):
            raise PyFunceble.exceptions.NoDownloadLinkGiven()

    def get_all_downtimes(self):
        """
        Provides all download times.

        :rtype: dict
        """

        return PyFunceble.helpers.Dict().from_json_file(self.downtimes_file.path)

    def get_downtime(self):
        """
        Provides the download timestamp of the current :code:`DOWN_TIME_INDEX`.

        :rtype: dict, None
        """

        if self.is_downtime_set():
            return self.all_downtimes[self.DOWNTIME_INDEX]

        return None

    def update_downtime(self):
        """
        Updates the current download time.
        """

        current_datetime = datetime.utcnow()

        self.all_downtimes[self.DOWNTIME_INDEX] = {
            "iso": current_datetime.isoformat(),
            "timestamp": current_datetime.timestamp(),
        }

    def save_downtimes(self):
        """
        Saves all downtimes.
        """

        PyFunceble.helpers.Dict(self.all_downtimes).to_json_file(
            self.downtimes_file.path
        )

    def is_downtime_set(self):
        """
        Checks if the downtime is set for the current :code:`DOWNTIME_INDEX`.

        :rtype: bool
        """

        return (
            self.DOWNTIME_INDEX in self.all_downtimes
            and self.all_downtimes[self.DOWNTIME_INDEX]
            and all(
                [
                    x in self.all_downtimes[self.DOWNTIME_INDEX]
                    and self.all_downtimes[self.DOWNTIME_INDEX][x]
                    for x in ["iso", "timestamp"]
                ]
            )
        )

    def is_last_download_expired(self):
        """
        Checks if the last downloaded file is expired.
        """

        if (
            not PyFunceble.helpers.File(self.destination).exists()
            or not self.is_downtime_set()
        ):
            return True

        last_download = datetime.fromtimestamp(self.get_downtime()["timestamp"])

        if (
            self.REDOWNLOAD_AFTER <= 0
            and (datetime.utcnow() - last_download).seconds < 3600
        ):
            return False

        if last_download + timedelta(days=self.REDOWNLOAD_AFTER) <= datetime.utcnow():
            return True

        return False

    def process(self):
        """
        Process the download and returns the downloaded text.

        :rtype: str, None
        """

        try:
            if self.is_last_download_expired() and PyFunceble.helpers.Download(
                self.download_link
            ).text(destination=self.destination):
                self.update_downtime()
                self.save_downtimes()
        except UnableToDownload as exception:
            print(f"{Fore.RED}{Style.BRIGHT}Could not downlod {str(exception)}")
            sys.exit(1)



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Dispatches between the diferrent cores (from the CLI.)

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import PyFunceble

from .execution_time import ExecutionTime


class Dispatcher:  # pylint: disable=too-few-public-methods, too-many-arguments
    """
    Dispatches to one of the other brain.

    .. note::
        We are talking about :code:`brain` because,
        each :code:`PyFunceble.core.*` is a unique brain
        which has it's own function ;-)

        Sometimes, 2 brains depends from the other one.
        But that's great because working together is always
        great ;-)

    :param preset:
        An instance of the configuration preset.
    :type preset: :py:class:`~PyFunceble.config.preset.Preset`
    :param list domain_or_ip:
        A list of subject to test.
    :param str file_path:
        A file path or a link.

        .. note::
            If a file path is given, we read and test its content.

        .. note::
            If a link is given, we download it's content and test
            it.

        .. warning::
            If given, we consider each line to be an adblock filter,
            a host file format or a subject to test.

    :param str url_file_path:
        A file path or a link.

        .. note::
            If a file path is given, we read and test its content.

        .. note::
            If a link is given, we download it's content and test
            it.

        .. warning::
            If given, we consider each line to be a URL to test.
    :param str url_to_test:
        A list of URL to test.
    """

    def __init__(
        self,
        preset,
        domain_or_ip=None,
        file_path=None,
        url_file_path=None,
        url_to_test=None,
        generate_results_only=False,
        generate_all_results_only=False,
    ):
        self.preset = preset
        self.preset.init_all()
        PyFunceble.LOGGER.debug(f"CONFIGURATION:\n{PyFunceble.CONFIGURATION}")

        if domain_or_ip or file_path or url_file_path or url_to_test:

            PyFunceble.core.CLI.logs_sharing()
            ExecutionTime("start")

            if generate_results_only or generate_all_results_only:
                PyFunceble.INTERN["do_not_clean"] = True

            if domain_or_ip:
                PyFunceble.core.Simple(domain_or_ip).domain()
            elif file_path:
                self.dispatch_file_test(
                    file_path, generate_results_only, generate_all_results_only
                )
            elif url_file_path:
                self.dispatch_url_file_test(
                    url_file_path, generate_results_only, generate_all_results_only
                )
            elif url_to_test:
                PyFunceble.core.Simple(url_to_test).url()

            PyFunceble.output.Percentage().log()

            PyFunceble.core.CLI.colorify_logo()
            ExecutionTime("stop")

            PyFunceble.core.CLI.stay_safe()
        else:
            PyFunceble.core.CLI.print_nothing_to_test()

    @classmethod
    def dispatch_file_test(
        cls, file_path, generate_results_only, generate_all_results_only
    ):
        """
        Dispatches to the brain in charge of file testing.

        :param str file_path: The file path to test.
        :param bool generate_results_only:
            Tell us to only regenerate from the data stored into the
            MariaDB/MySQL databases.
        :param bool generate_all_results_only:
            Tell us to only regenerate from the data stored into the
            MariaDB/MySQL databases.

            .. note::
                The difference with :code:`generate_results_only` is that
                it includes the retested which status didn't changed.
        """

        PyFunceble.output.Constructor()

        if PyFunceble.CONFIGURATION.multiprocess:
            if not generate_results_only and not generate_all_results_only:
                PyFunceble.core.Multiprocess(file_path, "domain").run_test()
            elif generate_all_results_only:
                PyFunceble.core.Multiprocess(file_path, "domain").generate_files(
                    include_entries_without_changes=True
                )
            else:
                PyFunceble.core.Multiprocess(file_path, "domain").generate_files()
        else:
            if not generate_results_only and not generate_all_results_only:
                PyFunceble.core.File(file_path, "domain").run_test()
            elif generate_all_results_only:
                PyFunceble.core.File(file_path, "domain").generate_files(
                    include_entries_without_changes=True
                )
            else:
                PyFunceble.core.File(file_path, "domain").generate_files()

    def dispatch_url_file_test(
        self, url_file_path, generate_results_only, generate_all_results_only
    ):
        """
        Dispatches to the brain in charge of url file testing.

        :param str url_file_path: The file to test.
        :param bool generate_results_only:
            Tell us to only regenerate from the data stored into the
            MariaDB/MySQL databases.
        """

        PyFunceble.output.Constructor()
        self.preset.file_url()

        if PyFunceble.CONFIGURATION.multiprocess:
            if not generate_results_only and not generate_all_results_only:
                PyFunceble.core.Multiprocess(url_file_path, "url").run_test()
            elif generate_all_results_only:
                PyFunceble.core.Multiprocess(url_file_path, "url").generate_files(
                    include_entries_without_changes=True
                )
            else:
                PyFunceble.core.Multiprocess(url_file_path, "url").generate_files()
        else:
            if not generate_results_only and not generate_all_results_only:
                PyFunceble.core.File(url_file_path, "url").run_test()
            elif generate_all_results_only:
                PyFunceble.core.File(url_file_path, "url").generate_files(
                    include_entries_without_changes=True
                )
            else:
                PyFunceble.core.File(url_file_path, "url").generate_files()



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the content/output generator.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from datetime import datetime
from os import sep as directory_separator

import PyFunceble


class Generate:  # pylint:disable=too-many-instance-attributes, too-many-arguments
    """
    Generate different sort of files.

    :param str subject: The subject we are working with.

    :param str subject_type: The type of the subject.

    :param str status: The catched status.

    :param str source: The source of the given status.

    :param str expiration_date:
        The expiration date of the domain (if catched).

    :param http_status_code: The HTTP status code.
    :type http_status_code: str|int

    :param str whois_server: The whois server.

    :param str filename:
        The name of the file we are testing.

    :param bool ip_validation:
        The IP validation check of the currently written subject.
    """

    def __init__(
        self,
        subject,
        subject_type,
        status,
        source=None,
        expiration_date=None,
        http_status_code=None,
        whois_server="Unknown",
        filename=None,
        ip_validation=False,
        end=False,
    ):
        # We share the subject.
        self.subject = subject
        # We share the subject type.
        self.subject_type = subject_type
        # We share the status
        self.status = status
        # We get the source.
        self.source = source
        # We share the file name.
        self.filename = filename
        # We share the IP validation.
        self.ip_validation = ip_validation
        # We share the end state.
        self.end = end

        if not http_status_code:
            self.status_code = PyFunceble.HTTP_CODE.not_found_default
        else:
            self.status_code = http_status_code

        if not whois_server:
            self.whois_server = "Unknown"
        else:
            # We share the whois server.
            self.whois_server = whois_server

        if not expiration_date:
            # The expiration date is not give or is empty.

            # We set the expiration date.
            self.expiration_date = "Unknown"
        else:
            # The expiration date is given.

            # We get the expiration date.
            self.expiration_date = expiration_date

        # We construct the output parent directory.
        self.output_parent_dir = (
            PyFunceble.OUTPUT_DIRECTORY + PyFunceble.OUTPUTS.parent_directory
        )

        user_agent = PyFunceble.engine.UserAgent().get()

        if user_agent:
            # The user-agent (from the configuration file) is not empty.

            # We initiate the header to use with our request.
            self.headers = {"User-Agent": user_agent}
        else:
            # The user-agent (from the configuration file) is empty.

            # We initiate an empty header to use with our request.
            self.headers = {}

    @property
    def file_production(self):
        """
        Check if we are allowed to produce a file based from the given
        information.

        :return:
            The state of the production.
            True: We do not produce file.
            False: We do produce file.
        :rtype: bool
        """

        if (
            "api_file_generation" in PyFunceble.CONFIGURATION
            and PyFunceble.CONFIGURATION.api_file_generation
        ):
            return bool(PyFunceble.CONFIGURATION.api_file_generation)

        if PyFunceble.CONFIGURATION.db_type in ["mariadb", "mysql"]:
            return self.end

        return not PyFunceble.CONFIGURATION.no_files

    def _analytic_host_file_directory(self):
        """
        Return the analytic directory to write depending of the matched
        status.
        """

        # We construct the path to the analytic directory.
        output_dir = (
            self.output_parent_dir + PyFunceble.OUTPUTS.analytic.directories.parent
        )

        if self.status.lower() in PyFunceble.STATUS.list.potentially_up:
            # The status is in the list of analytic up status.

            # We complete the output directory.
            output_dir += PyFunceble.OUTPUTS.analytic.directories.potentially_up
        elif self.status.lower() in PyFunceble.STATUS.list.potentially_down:
            # The status is in the list of analytic down status.

            # We complete the output directory.
            output_dir += PyFunceble.OUTPUTS.analytic.directories.potentially_down
        elif self.status.lower() in PyFunceble.STATUS.list.suspicious:
            # The status is in the list of analytic suspicious status.

            # We complete the output directory.
            output_dir += PyFunceble.OUTPUTS.analytic.directories.suspicious
        else:
            # The status is not in the list of analytic down or up status.

            # We complete the output directory.
            output_dir += PyFunceble.OUTPUTS.analytic.directories.up

        return output_dir

    def ___info_files_authorization(self):
        """
        Provide the authorization for the generation
        of info files.

        Basically here is what we check:

        * We are not testing as an imported module.

        and

        * The hosts file generation is activated.

        or

        * The plain list generation is activated.

        or

        * The "api_file_generation" was set into the CONFIGURATION.

        """

        return (
            self.file_production
            and self.subject_type.startswith("file_")
            and (
                PyFunceble.CONFIGURATION.generate_hosts
                or PyFunceble.CONFIGURATION.plain_list_domain
                or PyFunceble.CONFIGURATION.generate_json
            )
            or bool(PyFunceble.CONFIGURATION.api_file_generation)
        )

    def ___get_info_files_destinations(self, output_hosts, output_domains, output_json):
        """
        Given the output directory, this method return several paths.

        .. note::
            The given output directories have to be partially completed.

            Indeed, we only do :code:`output % final_location`.

        :return:
            The following paths:

            ::

                (
                    hosts_destination,
                    plain_destination,
                    json_destination,
                    splited_destination
                )

        :rtype: tuple
        """

        # We present the splited destination.
        splited_destination = None

        # We initiate the list of all analytic related statuses.
        http_list = []
        http_list.extend(PyFunceble.STATUS.list.potentially_up)
        http_list.extend(PyFunceble.STATUS.list.potentially_down)
        http_list.extend(PyFunceble.STATUS.list.http_active)
        http_list.extend(PyFunceble.STATUS.list.suspicious)

        if self.status.lower() in PyFunceble.STATUS.list.valid:
            # The status is in the list of valid list.

            # We complete the path to the hosts file.
            hosts_destination = output_hosts % PyFunceble.STATUS.official.valid

            # We complete the path to the plain list file.
            plain_destination = output_domains % PyFunceble.STATUS.official.valid

            # We complete the path to the json list file.
            json_destination = output_json % PyFunceble.STATUS.official.valid
        elif self.status.lower() in PyFunceble.STATUS.list.sane:
            # The status is in the list of sane list.

            # We complete the path to the hosts file.
            hosts_destination = output_hosts % PyFunceble.STATUS.official.sane

            # We complete the path to the plain list file.
            plain_destination = output_domains % PyFunceble.STATUS.official.sane

            # We complete the path to the json list file.
            json_destination = output_json % PyFunceble.STATUS.official.sane
        elif self.status.lower() in PyFunceble.STATUS.list.up:
            # The status is in the list of up list.

            # We complete the path to the hosts file.
            hosts_destination = output_hosts % PyFunceble.STATUS.official.up

            # We complete the path to the plain list file.
            plain_destination = output_domains % PyFunceble.STATUS.official.up

            # We complete the path to the json list file.
            json_destination = output_json % PyFunceble.STATUS.official.up
        elif self.status.lower() in PyFunceble.STATUS.list.malicious:
            # The status is in the list of malicious list.

            # We complete the path to the hosts file.
            hosts_destination = output_hosts % PyFunceble.STATUS.official.malicious

            # We complete the path to the plain list file.
            plain_destination = output_domains % PyFunceble.STATUS.official.malicious

            # We complete the path to the json list file.
            json_destination = output_json % PyFunceble.STATUS.official.malicious
        elif self.status.lower() in PyFunceble.STATUS.list.down:
            # The status is in the list of down list.

            # We complete the path to the hosts file.
            hosts_destination = output_hosts % PyFunceble.STATUS.official.down

            # We complete the path to the plain list file.
            plain_destination = output_domains % PyFunceble.STATUS.official.down

            # We complete the path to the json list file.
            json_destination = output_json % PyFunceble.STATUS.official.down
        elif self.status.lower() in PyFunceble.STATUS.list.invalid:
            # The status is in the list of invalid list.

            # We complete the path to the hosts file.
            hosts_destination = output_hosts % PyFunceble.STATUS.official.invalid

            # We complete the path to the plain list file.
            plain_destination = output_domains % PyFunceble.STATUS.official.invalid

            # We complete the path to the json list file.
            json_destination = output_json % PyFunceble.STATUS.official.invalid
        elif self.status.lower() in http_list:
            # The status is in the list of analytic status.

            # We construct the path to the analytic directory.
            output_dir = self._analytic_host_file_directory()

            if not output_dir.endswith(directory_separator):
                # The output directory does not ends with the directory separator.

                # We append the directory separator at the end of the output directory.
                output_dir += directory_separator

            # We initiate the hosts file path.
            hosts_destination = output_dir + PyFunceble.OUTPUTS.hosts.filename

            # We initiate the plain list file path.
            plain_destination = output_dir + PyFunceble.OUTPUTS.domains.filename

            # We complete the path to the json list file.
            json_destination = output_dir + PyFunceble.OUTPUTS.json.filename

            # We initiate the path to the http code file.
            # Note: We generate the http code file so that
            # we can have each domain in a file which is the
            # extracted http code.
            splited_destination = output_dir + str(self.status_code)
        elif self.status.lower().startswith("complements_"):
            # The status is in the list of complements status.

            # We convert the status to lower case.
            status = self.status.lower()
            # We get the status type.
            status_type = status[status.find("_") + 1 :]

            # We construct the path to the complements directory.
            output_dir = (
                self.output_parent_dir
                + PyFunceble.OUTPUTS.complements.directory
                + PyFunceble.STATUS.official[status_type]
            )

            if not output_dir.endswith(directory_separator):
                # The output directory does not ends with the directory separator.

                # We append the directory separator at the end of the output directory.
                output_dir += directory_separator

            # We initiate the hosts file path.
            hosts_destination = output_dir + PyFunceble.OUTPUTS.hosts.filename

            # We initiate the plain list file path.
            plain_destination = output_dir + PyFunceble.OUTPUTS.domains.filename

            # We complete the path to the json list file.
            json_destination = output_dir + PyFunceble.OUTPUTS.json.filename

        return (
            hosts_destination,
            plain_destination,
            json_destination,
            splited_destination,
        )

    # pylint: disable=inconsistent-return-statements,too-many-branches
    def info_files(self):
        """
        Generate the hosts file, the plain list, the JSON file and the splitted files.
        """

        if self.___info_files_authorization():
            # We initiate a variable which whill save the splited testination.
            splited_destination = ""

            # We partially initiate the path to the hosts file.
            output_hosts = (
                self.output_parent_dir
                + PyFunceble.OUTPUTS.hosts.directory
                + "%s"
                + directory_separator
                + PyFunceble.OUTPUTS.hosts.filename
            )

            # We partially initiate the path to the plain list file.
            output_domains = (
                self.output_parent_dir
                + PyFunceble.OUTPUTS.domains.directory
                + "%s"
                + directory_separator
                + PyFunceble.OUTPUTS.domains.filename
            )

            # We partially intiate the path to the json list file.
            output_json = (
                self.output_parent_dir
                + PyFunceble.OUTPUTS.json.directory
                + "%s"
                + directory_separator
                + PyFunceble.OUTPUTS.json.filename
            )

            if self.ip_validation:
                # The element is an IP.

                # We construct the output file.
                output_hosts = (
                    self.output_parent_dir
                    + PyFunceble.OUTPUTS.hosts.directory
                    + "%s"
                    + directory_separator
                    + PyFunceble.OUTPUTS.hosts.ip_filename
                )

            # We get the destination of the different files.
            (
                hosts_destination,
                plain_destination,
                json_destination,
                splited_destination,
            ) = self.___get_info_files_destinations(
                output_hosts, output_domains, output_json
            )

            if PyFunceble.CONFIGURATION.generate_hosts:
                # The hosts file generation is activated.

                # We generate/append the currently tested element in its
                # final location. (hosts file format)
                # We print on screen and on file.
                PyFunceble.output.Prints(
                    [PyFunceble.CONFIGURATION.custom_ip, self.subject],
                    "FullHosts",
                    hosts_destination,
                ).data()

            if PyFunceble.CONFIGURATION.plain_list_domain:
                # The plain list generation is activated.

                # We generate/append the currently tested element in its
                # final location. (the plain list format)
                # We print on file.
                PyFunceble.output.Prints(
                    [self.subject], "PlainDomain", plain_destination
                ).data()

            if PyFunceble.CONFIGURATION.split and splited_destination:
                # The splited list generation is activated.

                # We generate/append the currently tested element in its
                # final location. (the split list format)
                # We print on file.
                PyFunceble.output.Prints(
                    [self.subject], "PlainDomain", splited_destination
                ).data()

            if PyFunceble.CONFIGURATION.generate_json:
                # The json list generation is activated.

                # We generate/append the currently tested element in its
                # final location. (the json format)
                # We print on file.
                PyFunceble.output.Prints(
                    [self.subject], "JSON", json_destination
                ).data()

    def unified_file(self):
        """
        Generate unified file. Understand by that that we use an unified table
        instead of a separate table for each status which could result into a
        misunderstanding.
        """

        if (
            self.file_production
            and self.subject_type.startswith("file_")
            and PyFunceble.CONFIGURATION.unified
            and not PyFunceble.CONFIGURATION.split
        ):
            # * We are not testing as an imported module.
            # and
            # * The unified file generation is activated.

            # We construct the path of the unified file.
            output = self.output_parent_dir + PyFunceble.OUTPUTS.default_files.results

            if PyFunceble.CONFIGURATION.simple:
                to_print = [self.subject, self.status]

                PyFunceble.output.Prints(to_print, "Simple", output, True).data()
            elif PyFunceble.CONFIGURATION.less:
                # We have to print less information.

                if PyFunceble.HTTP_CODE.active:
                    # The http status code request is activated.

                    # We construct what we have to print.
                    to_print = [self.subject, self.status, self.status_code]
                else:
                    # The http status code request is not activated.

                    # We construct what we have to print.
                    to_print = [self.subject, self.status, self.source]

                # And we print the informations on file.
                PyFunceble.output.Prints(to_print, "Less", output, True).data()
            else:
                # The unified file generation is not activated.

                # We construct what we have to print.
                to_print = [
                    self.subject,
                    self.status,
                    self.expiration_date,
                    self.source,
                    self.status_code,
                    datetime.utcnow().isoformat(),
                ]

                # And we print the information on file.
                PyFunceble.output.Prints(to_print, "Generic_File", output, True).data()

    def complements_file(self):
        """
        Generate :code:`complements` files base on the current status.
        """

        if self.subject_type.startswith("file_"):
            # We are testing files.

            # We map the way we are going to work.
            status_map = {
                "up": "complements_UP",
                "down": "complements_DOWN",
                "invalid": "complements_INVALID",
                "valid": "complements_VALID",
            }

            for status, generate_status in status_map.items():
                # We loop through the list of status.

                if self.status.lower() in PyFunceble.STATUS.list[status]:
                    # The status is found.

                    # We generate the different files.
                    Generate(
                        self.subject,
                        self.subject_type,
                        generate_status,
                        source=self.source,
                        expiration_date=self.expiration_date,
                        http_status_code=self.status_code,
                        whois_server=self.whois_server,
                        filename=self.filename,
                        ip_validation=self.ip_validation,
                    ).info_files()

                    # We break the loop.
                    break

    def analytic_file(self, new_status, old_status=None):
        """
        Generate :code:`Analytic/*` files based on the given old and
        new statuses.

        :param str new_status: The new status of the domain.

        :param str old_status: The old status of the domain.
        """

        if not old_status:
            # The old status is not given.

            # We set the old status as the one given globally.
            old_status = self.status

        if self.subject_type.startswith("file_"):
            # We are testing files.

            # We map the way we are going to work with the status.
            status_map = {
                "up": "HTTP_Active",
                "potentially_up": "potentially_up",
                "suspicious": "suspicious",
            }

            # We keep a track of the map usage.
            map_used = False

            # We partially construct the path to the file to write/print.
            output = (
                self.output_parent_dir
                + PyFunceble.OUTPUTS.analytic.directories.parent
                + "%s%s"
            )

            for status, generate_status in status_map.items():
                # We loop through our map.

                if new_status.lower() in PyFunceble.STATUS.list[status]:
                    # The status is found into our map.

                    # We conmplete the output directory.
                    output = output % (
                        PyFunceble.OUTPUTS.analytic.directories[status],
                        PyFunceble.OUTPUTS.analytic.filenames[status],
                    )

                    # We generate the different file(s).
                    Generate(
                        self.subject,
                        self.subject_type,
                        generate_status,
                        source=self.source,
                        expiration_date=self.expiration_date,
                        http_status_code=self.status_code,
                        whois_server=self.whois_server,
                        filename=self.filename,
                        ip_validation=self.ip_validation,
                    ).info_files()

                    # We update the map usage.
                    map_used = True

                    # And we break the loop.
                    break

            if not map_used:
                # The map was not user, that means that
                # we are working with potentially inactive domains.

                # We complete the output directory.
                output = output % (
                    PyFunceble.OUTPUTS.analytic.directories.potentially_down,
                    PyFunceble.OUTPUTS.analytic.filenames.potentially_down,
                )

                # We generate the hosts files.
                Generate(
                    self.subject,
                    self.subject_type,
                    "potentially_down",
                    source=self.source,
                    expiration_date=self.expiration_date,
                    http_status_code=self.status_code,
                    whois_server=self.whois_server,
                    filename=self.filename,
                    ip_validation=self.ip_validation,
                ).info_files()

            # We print the information on file.
            PyFunceble.output.Prints(
                [
                    self.subject,
                    old_status,
                    self.status_code,
                    datetime.utcnow().isoformat(),
                ],
                "HTTP",
                output,
                True,
            ).data()

    def prints_status_file(self):  # pylint: disable=too-many-branches
        """
        Logic behind the printing (in file) when generating status file.
        """

        if (
            self.file_production
            and self.subject_type.startswith("file_")
            and PyFunceble.CONFIGURATION.split
        ):
            # We are testing a file.

            output = (
                self.output_parent_dir
                + PyFunceble.OUTPUTS.splited.directory
                + self.status
            )

            if PyFunceble.CONFIGURATION.simple:
                PyFunceble.output.Prints(
                    [self.subject, self.status], "Simple", output, True
                ).data()
            elif PyFunceble.CONFIGURATION.less:
                # We have to print less information.

                # We print the information on file.
                PyFunceble.output.Prints(
                    [self.subject, self.status, self.source], "Less", output, True
                ).data()
            elif PyFunceble.CONFIGURATION.split:
                # We have to split the information we print on file.

                if self.status.lower() in PyFunceble.STATUS.list.up:
                    # The status is in the list of up status.

                    if PyFunceble.HTTP_CODE.active:
                        # The http code extraction is activated.

                        # We initiate the data to print.
                        data_to_print = [
                            self.subject,
                            self.expiration_date,
                            self.source,
                            self.status_code,
                            datetime.utcnow().isoformat(),
                        ]
                    else:
                        # The http code extraction is not activated.

                        # We initiate the data to print.
                        data_to_print = [
                            self.subject,
                            self.expiration_date,
                            self.source,
                            datetime.utcnow().isoformat(),
                        ]

                    # We print the informations to print on file.
                    PyFunceble.output.Prints(
                        data_to_print, PyFunceble.STATUS.official.up, output, True
                    ).data()
                elif self.status.lower() in PyFunceble.STATUS.list.valid:
                    # The status is in the list of valid status.

                    # We initiate the data to print.
                    data_to_print = [
                        self.subject,
                        self.source,
                        datetime.utcnow().isoformat(),
                    ]

                    # We print the informations to print on file.
                    PyFunceble.output.Prints(
                        data_to_print, PyFunceble.STATUS.official.valid, output, True
                    ).data()
                elif self.status.lower() in PyFunceble.STATUS.list.sane:
                    # The status is in the list of sane status.

                    # We initiate the data to print.
                    data_to_print = [
                        self.subject,
                        self.source,
                        datetime.utcnow().isoformat(),
                    ]

                    # We print the informations to print on file.
                    PyFunceble.output.Prints(
                        data_to_print, PyFunceble.STATUS.official.sane, output, True
                    ).data()
                elif self.status.lower() in PyFunceble.STATUS.list.malicious:
                    # The status is in the list of malicious status.

                    # We initiate the data to print.
                    data_to_print = [
                        self.subject,
                        self.source,
                        datetime.utcnow().isoformat(),
                    ]

                    # We print the informations to print on file.
                    PyFunceble.output.Prints(
                        data_to_print,
                        PyFunceble.STATUS.official.malicious,
                        output,
                        True,
                    ).data()
                elif self.status.lower() in PyFunceble.STATUS.list.down:
                    # The status is in the list of down status.

                    if PyFunceble.HTTP_CODE.active:
                        # The http statuc code extraction is activated.

                        # We initiate the data to print.
                        data_to_print = [
                            self.subject,
                            self.whois_server,
                            self.status,
                            self.source,
                            self.status_code,
                            datetime.utcnow().isoformat(),
                        ]
                    else:
                        # The http status code extraction is not activated.

                        # We initate the data to print.
                        data_to_print = [
                            self.subject,
                            self.whois_server,
                            self.status,
                            self.source,
                            datetime.utcnow().isoformat(),
                        ]

                    # We print the information on file.
                    PyFunceble.output.Prints(
                        data_to_print, PyFunceble.STATUS.official.down, output, True
                    ).data()
                elif self.status.lower() in PyFunceble.STATUS.list.invalid:
                    # The status is in the list of invalid status.

                    if PyFunceble.HTTP_CODE.active:
                        # The http status code extraction is activated.

                        # We initiate the data to print.
                        data_to_print = [
                            self.subject,
                            self.source,
                            self.status_code,
                            datetime.utcnow().isoformat(),
                        ]
                    else:
                        # The http status code extraction is not activated.

                        # We initiate the data to print.
                        data_to_print = [
                            self.subject,
                            self.source,
                            datetime.utcnow().isoformat(),
                        ]

                    # We print the information to print on file.
                    PyFunceble.output.Prints(
                        data_to_print, PyFunceble.STATUS.official.invalid, output, True
                    ).data()

    def _prints_status_screen(self):
        """
        Logic behind the printing (on screen) when generating status file.
        """

        if not PyFunceble.CONFIGURATION.quiet:
            # The quiet mode is not activated.

            if PyFunceble.CONFIGURATION.simple:
                # We have to print simple infomation.

                # We initiate the data to print.
                to_print = [self.subject, self.status]

                PyFunceble.output.Prints(to_print, "Simple").data()
            elif PyFunceble.CONFIGURATION.less:
                # We have to print less information.

                # We initiate the data to print.
                to_print = [self.subject, self.status, self.status_code]

                if not PyFunceble.HTTP_CODE.active:
                    # The http status code is not activated.

                    # We replace the last element to print with
                    # the source.
                    to_print[-1] = self.source

                # We print the informations on screen.
                PyFunceble.output.Prints(to_print, "Less").data()
            else:
                # We have to print all informations on screen.

                if PyFunceble.HTTP_CODE.active:
                    # The http status code extraction is activated.

                    # We initiate the data to print.
                    data_to_print = [
                        self.subject,
                        self.status,
                        self.expiration_date,
                        self.source,
                        self.status_code,
                    ]
                else:
                    # The http status code extraction is not activated.

                    # We initiate the data to print.
                    data_to_print = [
                        self.subject,
                        self.status,
                        self.expiration_date,
                        self.source,
                        datetime.utcnow().isoformat(),
                    ]

                # We print the information on screen.
                PyFunceble.output.Prints(data_to_print, "Generic").data()
        elif "ci_initiated" in PyFunceble.INTERN or PyFunceble.CONFIGURATION.print_dots:
            # We are under a CI/CD environment.

            PyFunceble.LOGGER.info(f"Generation of output file for {self.subject!r}.")

            # We print a dot.
            print(".", end="")

    def status_file(
        self, exclude_file_generation=False
    ):  # pylint: disable=inconsistent-return-statements
        """
        Generate a file according to the domain status.

        :param bool exclude_file_generation:
            A shorthand to disable any file generation.
        """

        if exclude_file_generation:
            self.file_production = False

        # We generate the hosts file.
        self.info_files()

        if not self.end:
            # We print on screen if needed.
            self._prints_status_screen()

        # We increase the percentage count.
        PyFunceble.output.Percentage(self.status).count()

        if self.file_production:

            # We print or generate the  splitted files.
            self.prints_status_file()
            # We print or generate the unified files.
            self.unified_file()
